from toolmate import config, isServerAlive, print2
import os, sys

def main():
    if isServerAlive("127.0.0.1", config.llamacppToolModel_server_port):
        print2(f"A service is already running at 127.0.0.1:{config.llamacppToolModel_server_port}!")
        return None
    cmd = f"""{sys.executable} -m llama_cpp.server --port {config.llamacppToolModel_server_port} --model "{config.llamacppToolModel_model_path}" --verbose True --chat_format chatml --n_ctx {config.llamacppToolModel_n_ctx} --n_gpu_layers {config.llamacppToolModel_n_gpu_layers} --n_batch {config.llamacppToolModel_n_batch} {config.llamacppToolModel_additional_server_options}"""
    os.system(cmd)

def chat():
    if isServerAlive("127.0.0.1", config.llamacppChatModel_server_port):
        print2(f"A service is already running at 127.0.0.1:{config.llamacppChatModel_server_port}!")
        return None
    cmd = f"""{sys.executable} -m llama_cpp.server --port {config.llamacppChatModel_server_port} --model "{config.llamacppChatModel_model_path}" --verbose True --chat_format chatml --n_ctx {config.llamacppChatModel_n_ctx} --n_gpu_layers {config.llamacppChatModel_n_gpu_layers} --n_batch {config.llamacppChatModel_n_batch} {config.llamacppChatModel_additional_server_options}"""
    os.system(cmd)

def vision():
    if isServerAlive("127.0.0.1", config.llamacppVisionModel_server_port):
        print2(f"A service is already running at 127.0.0.1:{config.llamacppVisionModel_server_port}!")
        return None
    cmd = f"""{sys.executable} -m llama_cpp.server --port {config.llamacppVisionModel_server_port} --model "{config.llamacppVisionModel_model_path}" --clip_model_path {config.llamacppVisionModel_clip_model_path} --verbose True --chat_format llava-1-5 --n_ctx {config.llamacppVisionModel_n_ctx} --n_gpu_layers {config.llamacppVisionModel_n_gpu_layers} --n_batch {config.llamacppVisionModel_n_batch} {config.llamacppVisionModel_additional_server_options}"""
    os.system(cmd)

def custommain():
    if not config.customToolServer_ip.lower() in ("localhost", "127.0.0.1"):
        # a remote server
        return None
    elif config.customToolServer_command:
        if isServerAlive(config.customToolServer_ip, config.customToolServer_port):
            print2(f"A service is already running at {config.customToolServer_ip}:{config.customToolServer_port}!")
            return None
        os.system(config.customToolServer_command)
    else:
        main()

def customchat():
    if not config.customChatServer_ip.lower() in ("localhost", "127.0.0.1"):
        return None
    elif config.customChatServer_command:
        if isServerAlive(config.customChatServer_ip, config.customChatServer_port):
            print2(f"A service is already running at {config.customChatServer_ip}:{config.customChatServer_port}!")
            return None
        os.system(config.customChatServer_command)
    else:
        chat()

def customvision():
    if not config.customVisionServer_ip.lower() in ("localhost", "127.0.0.1"):
        return None
    elif config.customVisionServer_command:
        if isServerAlive(config.customVisionServer_ip, config.customVisionServer_port):
            print2(f"A service is already running at {config.customVisionServer_ip}:{config.customVisionServer_port}!")
            return None
        os.system(config.customVisionServer_command)
    else:
        vision()


"""
ToolMate AI Plugin - analyze audio file

analyze audio file

[TOOL_CALL]
"""

if not config.isTermux:

    from toolmate import config, showErrors, getGroqClient
    from toolmate import print1, print2, print3
    from toolmate.utils.sttLanguages import googleSpeeckToTextLanguages
    import os, whisper, io, shutil, subprocess
    import speech_recognition as sr
    from pydub import AudioSegment

    # Function method
    def analyze_audio_openai(function_args):
        from toolmate import config
        def check_file_format(file_path):
            # List of allowed file extensions
            allowed_extensions = ('.mp3', '.mp4', '.mpeg', '.mpga', '.m4a', '.wav', '.webm')
            # Getting the file extension
            _, file_extension = os.path.splitext(file_path)
            # Checking if the file extension is in the list of allowed extensions
            return True if file_extension.lower() in allowed_extensions else False

        audio_file = function_args.get("audio_filepath") # required
        #language = function_args.get("language") # required

        if audio_file and os.path.isfile(audio_file):
            if not check_file_format(audio_file):
                print3("This feature supports the following input file types only: '.mp3', '.mp4', '.mpeg', '.mpga', '.m4a', '.wav', '.webm'!")
                return ""
            elif os.path.getsize(audio_file) / (1024*1024) > 25:
                print3("Audio files are currently limited to 25 MB!")
                return ""
            try:
                with open(audio_file, "rb") as audio_file:
                    transcript = config.oai_client.audio.transcriptions.create(
                        model="whisper-1", 
                        file=audio_file, 
                        response_format="text"
                    )
                transcript = f"The transcript of the audio is: {transcript}"
                if config.developer:
                    print2(config.divider)
                    print3(transcript)
                    print2(config.divider)
                    print2("Answer to your enquiry:")
                return transcript
            except:
                showErrors()
            return ""

        return "[INVALID]"

    # Function Signature
    functionSignature = {
        "examples": [
            "analyze speech",
        ],
        "name": "analyze_audio_openai",
        "description": f'''Retrieve information from an audio with OpenAI''',
        "parameters": {
            "type": "object",
            "properties": {
                "audio_filepath": {
                    "type": "string",
                    "description": "Return the audio file path that I specified in my requests. Return an empty string '' if it is not specified.",
                },
            },
            "required": ["audio_filepath", "language"],
        },
    }

    # Integrate the signature and method into LetMeDoIt AI
    config.addFunctionCall(signature=functionSignature, method=analyze_audio_openai)


from toolmate import config, print1, print2, getCliOutput
if config.terminalEnableTermuxAPI:
    import pydoc
else:
    import pyperclip

# Tool: @copy_to_clipboard
def copy_to_clipboard(function_args):
    content = config.currentMessages[-1]["content"]
    pydoc.pipepager(content, cmd="termux-clipboard-set") if config.terminalEnableTermuxAPI else pyperclip.copy(content)
    message = "Copied!"
    print1(message)
    config.currentMessages[-1]["content"] = "Copy the following text to the system clipboard:\n\n```" + content + "\n```"
    config.currentMessages.append({"role": "assistant", "content": message})
    return ""
functionSignature = {
    "examples": [],
    "name": "copy_to_clipboard",
    "description": "Copy a given content to the system clipboard",
    "parameters": {
        "type": "object",
        "properties": {},
        "required": [],
    },
}
config.addFunctionCall(signature=functionSignature, method=copy_to_clipboard)

# Tool: @paste_from_clipboard
def paste_from_clipboard(function_args):
    content = config.currentMessages[-1]["content"]
    clipboardText = getCliOutput("termux-clipboard-get") if config.terminalEnableTermuxAPI else pyperclip.paste()
    print2("\n```clipboard")
    print1(f"{clipboardText}")
    print2("```")
    if not content.strip() == "Retrieve the clipboard text":
        config.currentMessages.append({"role": "assistant", "content": clipboardText})
        return ""
    config.currentMessages[-1]["content"] = f"{content}\n\n{clipboardText}"
    return "[INVALID]" # this take to generate a chat response based on the user input, together with the clipboard content
functionSignature = {
    "examples": [],
    "name": "paste_from_clipboard",
    "description": "Retrieve the text content from the system clipboard and paste",
    "parameters": {
        "type": "object",
        "properties": {},
        "required": [],
    },
}
config.addFunctionCall(signature=functionSignature, method=paste_from_clipboard)

'''
Notes about pbcopy and pbpaste alternatives:

alias pbcopy='xclip -selection clipboard'
alias pbpaste='xclip -selection clipboard -o'

alias pbcopy='xsel --clipboard --input'
alias pbpaste='xsel --clipboard --output'
'''


"""
ToolMate AI Plugin - search weather info

search for weather information

[TOOL_CALL]
"""

if not config.openweathermapApi or config.openweathermapApi == "toolmate":
    config.changeOpenweathermapApi()

if config.openweathermapApi:
    from toolmate import config
    from toolmate import print1, print3, getOpenweathermapApi_key
    from toolmate.utils.python_utils import PythonUtil
    import json

    def search_weather_info(function_args):
        code = function_args.get("code") # required
        information = PythonUtil.showAndExecutePythonCode(code)
        if information:
            return json.loads(information)["information"]
        return "Not found!"

    functionSignature = {
        "examples": [
            "current weather",
            "search weather",
        ],
        "name": "search_weather_info",
        "description": f'''Answer a query about weather''',
        "parameters": {
            "type": "object",
            "properties": {
                "code": {
                    "type": "string",
                    "description": f"""Generate python code that use my OpenWeatherMap API key '{getOpenweathermapApi_key()}' to resolve my request. Use Celsius as the unit for temperature.""",
                },
            },
            "required": ["code"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=search_weather_info)
else:
    print1("To use plugin 'search weather info', you need to set up an OpenWeatherMap API key first.")
    print3("Read: https://github.com/eliranwong/letmedoit/wiki/OpenWeatherMap-API-Setup")


"""
ToolMate AI Plugin - ask PaLM 2

Ask Google PaLM 2 for information

[TOOL_CALL]
"""

if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs and not config.isTermux:

    import vertexai
    from vertexai.generative_models._generative_models import (
        HarmCategory,
        HarmBlockThreshold,
    )
    from toolmate import config
    import vertexai
    from vertexai.language_models import ChatModel, ChatMessage

    def ask_palm2(function_args):
        config.stopSpinning()
        query = function_args.get("query") # required
        config.currentMessages[-1] = {"role": "user", "content": query}

        model = ChatModel.from_pretrained("chat-bison-32k")
        # https://cloud.google.com/vertex-ai/docs/generative-ai/model-reference/text-chat
        parameters = {
            "temperature": config.llmTemperature,  # Temperature controls the degree of randomness in token selection; 0.0–1.0; Default: 0.0
            "max_output_tokens": 2048,  # Token limit determines the maximum amount of text output; 1–2048; Default: 1024
            "top_p": 0.95,  # Tokens are selected from most probable to least until the sum of their probabilities equals the top_p value; 0.0–1.0; Default: 0.95
            "top_k": 40,  # A top_k of 1 means the selected token is the most probable among all tokens; 1-40; Default: 40
        }
        history = []
        user = True
        for i in config.currentMessages:
            if i["role"] == "user" if user else "assistant":
                history.append(ChatMessage(content=i["content"], author="user" if user else "model"))
                user = not user
        if history and history[-1].author == "user":
            history = history[:-1]
        elif not history:
            history = None
        chat = model.start_chat(
            context=config.systemMessage_palm2,
            message_history=history,
            #examples=[
            #    InputOutputTextPair(
            #        input_text="How many moons does Mars have?",
            #        output_text="The planet Mars has two moons, Phobos and Deimos.",
            #    ),
            #],
        )

        completion = chat.send_message_streaming(query, **parameters)
        config.toolmate.streamCompletion(completion, openai=False)
        return ""

    functionSignature = {
        "examples": [
            "Ask PaLM",
        ],
        "name": "ask_palm2",
        "description": "Ask PaLM 2 to chat or provide information",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The request in detail, including any supplementary information",
                },
            },
            "required": ["query"],
        },
    }

    # initiation
    vertexai.init()

    # Note: BLOCK_NONE is not allowed
    config.gemini_safety_settings={
        HarmCategory.HARM_CATEGORY_UNSPECIFIED: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
    }
    config.addFunctionCall(signature=functionSignature, method=ask_palm2)
    config.inputSuggestions.append("Ask PaLM 2: ")
else:
    print("Plugin `ask palm2` not enabled! Read setup at https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md")


"""
ToolMate AI Plugin - remove image background

Remove image background

[TOOL_CALL]
"""


if not config.isTermux:

    from toolmate import config, is_valid_image_file
    from toolmate import print2, print3
    from toolmate.utils.python_utils import PythonUtil
    import os, json, rembg, shutil


    def remove_image_background(function_args):
        files = function_args.get("filepath") # required
        if isinstance(files, str):
            if not files.startswith("["):
                files = f'["{files}"]'
            files = eval(files)
        if not files:
            return "[INVALID]"

        filesCopy = files[:]
        for item in filesCopy:
            if os.path.isdir(item):
                for root, _, allfiles in os.walk(item):
                    for file in allfiles:
                        file_path = os.path.join(root, file)
                        files.append(file_path)
                files.remove(item)

        for input_path in files:
            if is_valid_image_file(input_path):
                output_path = f"{input_path}_no_bg.png"
                with open(input_path, 'rb') as i:
                    with open(output_path, 'wb') as o:
                        print3(f"Reading image file: {input_path}")
                        img = rembg.remove(i.read())
                        o.write(img)
                        print3(f"File saved: {output_path}")
            else:
                print2(f"'{input_path}' is not an image file!")

    def remove_image_background2(function_args):
        code = function_args.get("code") # required
        information = PythonUtil.showAndExecutePythonCode(code)
        if information:
            filepath = json.loads(information)["information"]
            if os.path.isfile(filepath):
                if shutil.which(config.open):
                    try:
                        os.system(f'''{config.open} "{filepath}"''')
                    except:
                        pass
                config.toolTextOutput = f"Image saved: {filepath}"
                print3(config.toolTextOutput)
        return ""

    functionSignature = {
        "examples": [
            "remove image background",
        ],
        "name": "remove_image_background",
        "description": f'''Remove image background''',
        "parameters": {
            "type": "object",
            "properties": {
                "filepath": {
                    "type": "string",
                    "description": """Return a list of image paths, e.g. '["image1.png", "/tmp/image2.png"]'. Return '[]' if image path is not provided.""",
                },
            },
            "required": ["filepath"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=remove_image_background)


"""
ToolMate AI Plugin - help

Retrieve information from the documentation regarding how to use ToolMate AI

[TOOL_CALL]
"""


if not config.isTermux:

    from toolmate import config, getHelpCollection, query_vectors, getRagPrompt, print2
    from toolmate.utils.call_llm import CallLLM
    import os


    def help(function_args):
        query = function_args.get("query") # required
        if query.strip():
            collection = getHelpCollection()
            retrievedContext = query_vectors(collection, query, n=config.rag_closestMatches)["documents"]
            ragPrompt = getRagPrompt(query, retrievedContext)

            messages = config.currentMessages[:-1] + [{"role": "user", "content" : ragPrompt}]
            completion = CallLLM.regularCall(messages)
            print2(config.divider)
            config.toolmate.streamCompletion(completion)
            print2(config.divider)
        else:
            config.currentMessages = config.currentMessages[:-1]
        return ""

    functionSignature = {
        "examples": [
            "how to use ToolMate AI",
        ],
        "name": "help",
        "description": "Retrieve information from the documentation regarding how to use ToolMate AI",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "Detailed queries about how to use ToolMate AI",
                },
            },
            "required": ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=help)


from toolmate import config, getDeviceInfo, restartApp, count_tokens_from_messages, isCommandInstalled, getCpuThreads
from toolmate import print1, print2, print3, count_tokens_from_functions, voiceTyping, tokenLimits
import pydoc, textwrap, re
from prompt_toolkit import prompt
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.styles import Style
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings, ConditionalKeyBindings
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from toolmate.utils.prompt_shared_key_bindings import prompt_shared_key_bindings
from toolmate.utils.prompt_multiline_shared_key_bindings import prompt_multiline_shared_key_bindings
from toolmate.utils.prompt_validator import NumberValidator
import tiktoken

class Prompts:

    def __init__(self):
        config.multilineInput = False
        config.clipboard = PyperclipClipboard()
        config.promptStyle1 = self.promptStyle1 = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor1,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor1,
        })
        config.promptStyle2 = self.promptStyle2 = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        self.inputIndicator = [
            ("class:indicator", ">>> "),
        ]
        self.shareKeyBindings()
        config.showKeyBindings = self.showKeyBindings
        config.terminalColors = {
            "ansidefault": "ansidefault",
            "ansiblack": "ansiwhite",
            "ansired": "ansibrightred",
            "ansigreen": "ansibrightgreen",
            "ansiyellow": "ansibrightyellow",
            "ansiblue": "ansibrightblue",
            "ansimagenta": "ansibrightmagenta",
            "ansicyan": "ansibrightcyan",
            "ansigray": "ansibrightblack",
            "ansiwhite": "ansiblack",
            "ansibrightred": "ansired",
            "ansibrightgreen": "ansigreen",
            "ansibrightyellow": "ansiyellow",
            "ansibrightblue": "ansiblue",
            "ansibrightmagenta": "ansimagenta",
            "ansibrightcyan": "ansicyan",
            "ansibrightblack": "ansigray",
        }

    def getToolBar(self, multiline=False):
        if multiline:
            return f""" {str(config.hotkey_exit).replace("'", "")} exit [escape+enter] send """
        return f""" {str(config.hotkey_exit).replace("'", "")} exit [enter] send """

    def shareKeyBindings(self):
        this_key_bindings = KeyBindings()

        @this_key_bindings.add(*config.hotkey_voice_entry)
        def _(event):
            if config.pyaudioInstalled:
                buffer = event.app.current_buffer
                buffer.text = f"{buffer.text}{' ' if buffer.text else ''}{voiceTyping()}"
                if config.voiceTypingAutoComplete:
                    buffer.validate_and_handle()
                else:
                    buffer.cursor_position = buffer.cursor_position + buffer.document.get_end_of_line_position()
            else:
                run_in_terminal(lambda: print2("Install PyAudio first to enable voice recognition!"))
        @this_key_bindings.add(*config.hotkey_voice_recognition_config)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".speechrecognition"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_voice_generation_config)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".speechgeneration"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_select_plugins)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".plugins"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_insert_favorite_string)
        def _(event):
            buffer = event.app.current_buffer
            buffer.text += config.favorite_string
            buffer.cursor_position = len(buffer.text)
        """@this_key_bindings.add(*config.hotkey_list_directory_content)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".content"
            buffer.validate_and_handle()"""
        @this_key_bindings.add(*config.hotkey_exit)
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = config.exit_entry
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_insert_filepath) # add path
        def _(event):
            buffer = event.app.current_buffer
            config.addPathAt = buffer.cursor_position
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_new)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".new"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_open_chat_records)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".open"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_open_previous_conversation)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".last"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_read_conversation)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".read"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_export)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".export"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_display_device_info)
        def _(_):
            run_in_terminal(lambda: print(getDeviceInfo(True)))
        @this_key_bindings.add(*config.hotkey_count_tokens)
        def _(event):
            try:
                encoding = tiktoken.get_encoding("cl100k_base")
                currentInput = event.app.current_buffer.text
                #availableFunctionTokens = count_tokens_from_functions(config.toolFunctionSchemas)
                currentInputTokens = len(encoding.encode(config.addPredefinedContext(currentInput)))
                loadedMessageTokens = count_tokens_from_messages(config.currentMessages)
                #selectedModelLimit = tokenLimits[config.chatGPTApiModel]
                #estimatedAvailableTokens = selectedModelLimit - availableFunctionTokens - loadedMessageTokens - currentInputTokens

                content = f"""# Token Count
Loaded messages: {loadedMessageTokens}
Current input: {currentInputTokens}
{config.divider}
"""
            except:
                content = "Required package 'tiktoken' not found!"
            run_in_terminal(lambda: print1(content))
        @this_key_bindings.add(*config.hotkey_launch_pager_view)
        def _(_):
            config.launchPager()
        @this_key_bindings.add(*config.hotkey_toggle_developer_mode)
        def _(_):
            config.developer = not config.developer
            config.saveConfig()
            run_in_terminal(lambda: print3(f"Developer mode: {'enabled' if config.developer else 'disabled'}!"))
        @this_key_bindings.add(*config.hotkey_toggle_multiline_entry)
        def _(_):
            config.toggleMultiline()
        '''
        @this_key_bindings.add(*config.hotkey_select_context)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".context"
            buffer.validate_and_handle()
        '''
        @this_key_bindings.add("escape", "!")
        @this_key_bindings.add(*config.hotkey_launch_system_prompt)
        def _(event):
            buffer = event.app.current_buffer
            config.defaultEntry = buffer.text
            buffer.text = ".system"
            buffer.validate_and_handle()
        @this_key_bindings.add(*config.hotkey_display_key_combo)
        def _(_):
            run_in_terminal(self.showKeyBindings)
        @this_key_bindings.add(*config.hotkey_toggle_input_audio)
        def _(_):
            if config.tts:
                config.ttsInput = not config.ttsInput
                config.saveConfig()
                run_in_terminal(lambda: print3(f"Input Audio: '{'enabled' if config.ttsInput else 'disabled'}'!"))
        @this_key_bindings.add(*config.hotkey_toggle_response_audio)
        def _(_):
            if config.tts:
                config.ttsOutput = not config.ttsOutput
                config.saveConfig()
                run_in_terminal(lambda: print3(f"Response Audio: '{'enabled' if config.ttsOutput else 'disabled'}'!"))
        #@this_key_bindings.add(*config.hotkey_restart_app)
        #def _(_):
        #    print(f"Restarting {config.toolMateAIName} ...")
        #    restartApp()
        @this_key_bindings.add(*config.hotkey_toggle_input_improvement)
        def _(_):
            config.improveInputEntry = not config.improveInputEntry
            config.saveConfig()
            run_in_terminal(lambda: print3(f"Improved Writing Display: '{'enabled' if config.improveInputEntry else 'disabled'}'!"))
        @this_key_bindings.add(*config.hotkey_toggle_word_wrap)
        def _(_):
            config.wrapWords = not config.wrapWords
            config.saveConfig()
            run_in_terminal(lambda: print3(f"Word Wrap: '{'enabled' if config.wrapWords else 'disabled'}'!"))
        @this_key_bindings.add(*config.hotkey_toggle_mouse_support)
        def _(_):
            config.mouseSupport = not config.mouseSupport
            config.saveConfig()
            run_in_terminal(lambda: print3(f"Entry Mouse Support: '{'enabled' if config.mouseSupport else 'disabled'}'!"))
        # edit the last response in built-in or custom text editor
        @this_key_bindings.add(*config.hotkey_edit_last_response)
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = ".editresponse"
            buffer.validate_and_handle()

        conditional_prompt_multiline_shared_key_bindings = ConditionalKeyBindings(
            key_bindings=prompt_multiline_shared_key_bindings,
            filter=Condition(lambda: config.multilineInput),
        )
        self.prompt_shared_key_bindings = merge_key_bindings([
            this_key_bindings,
            conditional_prompt_multiline_shared_key_bindings,
            prompt_shared_key_bindings,
        ])
    def showKeyBindings(self):
        def transformKey(entry):
            if not entry.startswith("["):
                entry = f"[{entry}]"
            entry = entry.replace("'", "")
            entry = entry.replace("[c-", "[ctrl, ")
            return entry
        bindings = {
            "[enter]": "complete entry",
            str(config.hotkey_insert_newline): "new line",
            str(config.hotkey_exit): "quit / exit current feature",
            str(config.hotkey_cancel): "cancel",
            "[c-a]": "select / unselect all",
            "[c-c]": "copy [w/ mouse support]",
            "[c-v]": "paste [w/ mouse support]",
            "[c-x]": "cut [w/ mouse support]",
            "[c-d]": "forward delete",
            "[c-h]": "backspace",
            "[shift, tab]": f"insert '{config.terminalEditorTabText}' (configurable)",
            "[escape, b]": "move cursor to line beginning",
            "[escape, e]": "move cursor to line end",
            "[escape, a]": "move cursor to entry beginning",
            "[escape, z]": "move cursor to entry end",
            "[c-r]": "reverse-i-search",
            str(config.hotkey_new): "new conversation",
            str(config.hotkey_open_chat_records): "open conversation",
            str(config.hotkey_open_previous_conversation): "open previous conversation",
            str(config.hotkey_read_conversation): "read current conversation",
            str(config.hotkey_edit_last_response): "edit current conversation",
            str(config.hotkey_export): "export conversation",
            #str(config.hotkey_select_context): "change predefined context",
            #str(config.hotkey_remove_context_temporarily): "remove context temporarily",
            str(config.hotkey_launch_pager_view): "launch pager view",
            str(config.hotkey_display_key_combo): "show key bindings",
            str(config.hotkey_voice_entry): "activate voice typing",
            str(config.hotkey_voice_recognition_config): "change voice typing configs",
            str(config.hotkey_toggle_input_audio): "toggle input audio",
            str(config.hotkey_toggle_response_audio): "toggle response audio",
            str(config.hotkey_toggle_multiline_entry): "toggle multi-line entry",
            str(config.hotkey_toggle_word_wrap): "toggle word wrap",
            str(config.hotkey_insert_favorite_string): "insert my favourite string",
            str(config.hotkey_insert_filepath): "insert a file or folder path",
            str(config.hotkey_display_device_info): "display device information",
            str(config.hotkey_count_tokens): "count current message tokens",
            str(config.hotkey_toggle_input_improvement): "toggle improved writing feature",
            str(config.hotkey_toggle_mouse_support): "toggle mouse support",
            str(config.hotkey_launch_system_prompt): "system command prompt",
            str(config.hotkey_swap_text_brightness): "swap text brightness",
            str(config.hotkey_toggle_developer_mode): "swap developer mode",
            str(config.hotkey_restart_app): "restart toolmate",
        }
        textEditor = config.customTextEditor.split(" ", 1)[0]
        bindings[str(config.hotkey_edit_current_entry)] = f"""edit current input with '{config.customTextEditor if textEditor and isCommandInstalled(textEditor) else "eTextEdit"}'"""
        #bindings[str(config.hotkey_edit_last_response)] = f"""edit the previous response with '{config.customTextEditor if textEditor and isCommandInstalled(textEditor) else "eTextEdit"}'"""
        multilineBindings = {
            "[enter]": "new line",
            "[escape, enter]": "complete entry",
            "[escape, 1]": "go up 10 lines",
            "[escape, 2]": "go up 20 lines",
            "[escape, 3]": "go up 30 lines",
            "[escape, 4]": "go up 40 lines",
            "[escape, 5]": "go up 50 lines",
            "[escape, 6]": "go up 60 lines",
            "[escape, 7]": "go up 70 lines",
            "[escape, 8]": "go up 80 lines",
            "[escape, 9]": "go up 90 lines",
            "[escape, 0]": "go up 100 lines",
            "[f1]": "go down 10 lines",
            "[f2]": "go down 20 lines",
            "[f3]": "go down 30 lines",
            "[f4]": "go down 40 lines",
            "[f5]": "go down 50 lines",
            "[f6]": "go down 60 lines",
            "[f7]": "go down 70 lines",
            "[f8]": "go down 80 lines",
            "[f9]": "go down 90 lines",
            "[f10]": "go down 100 lines",
            "[c-u]": f"go up '{config.terminalEditorScrollLineCount}' lines (configurable)",
            "[c-j]": f"go down '{config.terminalEditorScrollLineCount}' lines (configurable)",
        }

        keyHelp = f"{config.divider}\n\n# Tools\n\nEnter `@` to read brief descriptions of all enabled tools.\n"
        
        keyHelp += f"\n{config.divider}\n\n"
        keyHelp += config.actionHelp

        keyHelp += f"\n{config.divider}\n\n"
        keyHelp += "# Key Bindings\n"
        keyHelp += "[BLANK]: launch action menu\n"
        for key, value in bindings.items():
            key = transformKey(key)
            keyHelp += f"{key} {value}\n"
        keyHelp += "\n## Key Bindings\n(for multiline entry only)\n"
        for key, value in multilineBindings.items():
            key = transformKey(key)
            keyHelp += f"{key} {value}\n"

        keyHelp += f"\n{config.divider}\n"

        if isCommandInstalled("less"):
            pydoc.pipepager(f"To close this help page, press 'q'\n\n{keyHelp}\nTo close this help page, press 'q'", cmd='less -R')
        else:
            print(keyHelp)

    def simplePrompt(self, numberOnly=False, validator=None, inputIndicator="", default="", accept_default=False, completer=None, complete_in_thread=None, promptSession=None, style=None, is_password=False, bottom_toolbar=None):
        config.selectAll = False
        inputPrompt = promptSession.prompt if promptSession is not None else prompt
        if not inputIndicator:
            inputIndicator = self.inputIndicator
        if numberOnly:
            validator = NumberValidator()
        userInput = inputPrompt(
            inputIndicator,
            key_bindings=self.prompt_shared_key_bindings,
            bottom_toolbar=self.getToolBar(config.multilineInput) if bottom_toolbar is None else bottom_toolbar,
            #enable_system_prompt=True,
            swap_light_and_dark_colors=Condition(lambda: not config.terminalResourceLinkColor.startswith("ansibright")),
            style=self.promptStyle1 if style is None else style,
            validator=validator,
            multiline=Condition(lambda: config.multilineInput),
            default=default,
            accept_default=accept_default,
            completer=completer,
            complete_in_thread=complete_in_thread,
            is_password=is_password,
            mouse_support=Condition(lambda: config.mouseSupport),
            clipboard=config.clipboard,
        )
        userInput = textwrap.dedent(userInput) # dedent to work with code block
        return userInput if hasattr(config, "addPathAt") and config.addPathAt else userInput.strip()



from toolmate.utils.call_llm import CallLLM

from toolmate import config, getPythonFunctionResponse, fineTunePythonCode, getPygmentsStyle, showErrors
from toolmate import print1, extractPythonCode, displayPythonCode
import json, pygments
from pygments.lexers.python import PythonLexer
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import PygmentsTokens


class PythonUtil:

    @staticmethod
    def showAndExecutePythonCode(code):
        validCode = extractPythonCode(code)
        if validCode:
            config.stopSpinning()
            if config.developer or config.codeDisplay:
                displayPythonCode(validCode)
            refinedCode = fineTunePythonCode(validCode)
            information = PythonUtil.executePythonCode(refinedCode)
            return information
        return json.dumps({"information": code})

    @staticmethod
    def executePythonCode(code):
        try:
            exec(code, globals())
            pythonFunctionResponse = getPythonFunctionResponse(code)
        except:
            trace = showErrors()
            print1(config.divider)
            if config.max_consecutive_auto_correction > 0:
                return CallLLM.autoCorrectPythonCode(code, trace)
            else:
                return "[INVALID]"
        if not pythonFunctionResponse:
            return ""
        return json.dumps({"information": pythonFunctionResponse})



"""
ToolMate AI Plugin - convert simplified Chinese into traditional Chinese

Convert simplified Chinese into traditional Chinese in text output
"""

if not config.isTermux:

    try:
        from opencc import OpenCC
    except:
        from toolmate import installPipPackage
        installPipPackage(f"--upgrade opencc")

    from toolmate import config
    from opencc import OpenCC

    def convertToTraditionalChinese(text):
        if text:
            return OpenCC('s2t').convert(text)
        else:
            return text

    config.outputTransformers.append(convertToTraditionalChinese)


from toolmate import config
from toolmate.utils.call_llm import CallLLM
from toolmate.utils.tool_plugins import Plugins
from toolmate.gui.worker import Worker
import json, getpass

from PySide6.QtPrintSupport import QPrinter, QPrintDialog
from PySide6.QtCore import Qt, QThread, QRegularExpression, QRunnable, Slot, Signal, QObject, QThreadPool
from PySide6.QtGui import QStandardItemModel, QStandardItem, QGuiApplication, QAction, QIcon, QFontMetrics, QTextDocument
from PySide6.QtWidgets import QCompleter, QMenu, QSystemTrayIcon, QApplication, QMainWindow, QWidget, QDialog, QFileDialog, QDialogButtonBox, QFormLayout, QLabel, QMessageBox, QCheckBox, QPlainTextEdit, QProgressBar, QPushButton, QListView, QHBoxLayout, QVBoxLayout, QLineEdit, QSplitter, QComboBox

class CentralWidget(QWidget):

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        # set up interface
        self.setupUI()
        # set up variables
        self.setupVariables()
    
    def setupVariables(self):
        # username
        self.user = getpass.getuser().split(" ")[0].capitalize()
        self.assistant = config.toolMateAIName.split(" ")[0]
        # load LetMeDoIt AI system message
        config.currentMessages=CallLLM.resetMessages()
        # thread pool
        self.threadpool = QThreadPool()
        # scroll bar
        self.contentScrollBar = self.contentView.verticalScrollBar()

    def setupUI(self):
        # a layout with left and right columns and a splitter placed between them
        layout000 = QHBoxLayout()
        self.setLayout(layout000)
        
        widgetLt = QWidget()
        layout000Lt = QVBoxLayout()
        widgetLt.setLayout(layout000Lt)
        widgetRt = QWidget()
        layout000Rt = QVBoxLayout()
        widgetRt.setLayout(layout000Rt)
        
        splitter = QSplitter(Qt.Horizontal, self)
        splitter.addWidget(widgetLt)
        splitter.addWidget(widgetRt)
        layout000.addWidget(splitter)

        # add widgets to left column later
        widgetLt.hide()

        # widgets
        # user input
        self.userInput = QLineEdit()
        completer = QCompleter(config.inputSuggestions)
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        completer.setCompletionMode(QCompleter.PopupCompletion)
        completer.setFilterMode(Qt.MatchContains)
        self.userInput.setCompleter(completer)
        self.userInput.setPlaceholderText("Enter your request here ...")
        self.userInput.mousePressEvent = lambda _ : self.userInput.selectAll()
        self.userInput.setClearButtonEnabled(True)
        # content view
        self.contentView = QPlainTextEdit()
        self.contentView.setReadOnly(True)
        # progress bar
        self.progressBar = QProgressBar()
        self.progressBar.setRange(0, 0) # Set the progress bar to use an indeterminate progress indicator
        
        # update layout
        layout000Rt.addWidget(self.userInput)
        layout000Rt.addWidget(self.contentView)
        layout000Rt.addWidget(self.progressBar)
        self.progressBar.hide()

        # Connections
        self.userInput.returnPressed.connect(self.submit)

    def submit(self):
        if request := self.userInput.text().strip():
            self.userInput.setDisabled(True)
            self.addContent(request)

        self.progressBar.show()
        config.toolmate.runMultipleActions(request, True)

    def processResponse(self):
        self.userInput.setText("")
        self.userInput.setEnabled(True)
        self.progressBar.hide()
        self.userInput.setFocus()

    def streamResponse(self, content):
        self.contentView.insertPlainText(content)
        self.contentScrollBar.setValue(self.contentScrollBar.maximum())

    def addContent(self, newContent, user=True) -> None:
        content = self.contentView.toPlainText()
        if content:
            self.contentView.insertPlainText("\n\n")
        if user:
            self.contentView.insertPlainText(f"[{self.user}] {newContent}\n\n[{self.assistant}] ")
        else:
            self.contentView.insertPlainText(newContent)

class DesktopAssistant(QMainWindow):
    def __init__(self, standalone=False) -> None:
        super().__init__()
        # check if running standalone
        self.standalone = standalone
        # set title
        self.setWindowTitle(config.toolMateAIName)
        # set variables
        #self.setupVariables()
        # run plugins
        Plugins.runPlugins()
        # gui
        self.initUI()
        # shortcuts
        self.processResponse = self.centralWidget.processResponse
        self.streamResponse = self.centralWidget.streamResponse

    def closeEvent(self, event):
        if self.standalone:
            event.accept()
        else:
            # hiding it, instead of closing it, to save from reloading time
            event.ignore()
            self.hide()

    def initUI(self):
        config.conversationStarted = False
        self.centralWidget = CentralWidget(self)
        self.setCentralWidget(self.centralWidget)
        self.resize(config.desktopAssistantWidth, config.desktopAssistantHeight)
        #
        self.createMenubar()

    def resizeEvent(self, event):
        size = event.size()
        config.desktopAssistantWidth = size.width()
        config.desktopAssistantHeight = size.height()
        config.saveConfig()
    
    def printTextOutput(self, output):
        self.centralWidget.addContent(output, False)

    def createMenubar(self):
        # Create a menu bar
        menubar = self.menuBar()

        # Create a File menu and add it to the menu bar
        file_menu = menubar.addMenu("Chat")

        new_action = QAction("New", self)
        new_action.setShortcut("Ctrl+N")
        new_action.triggered.connect(self.newConversation)
        file_menu.addAction(new_action)

        """new_action = QAction("Open", self)
        new_action.setShortcut("Ctrl+O")
        new_action.triggered.connect(self.openConversation)
        file_menu.addAction(new_action)

        new_action = QAction("Save", self)
        new_action.setShortcut("Ctrl+S")
        new_action.triggered.connect(self.saveConversation)
        file_menu.addAction(new_action)

        new_action = QAction("Save As...", self)
        new_action.triggered.connect(self.saveAsConversation)
        file_menu.addAction(new_action)

        new_action = QAction("Export", self)
        new_action.triggered.connect(self.exportConversation)
        file_menu.addAction(new_action)

        file_menu.addSeparator()"""

    def newConversation(self):
        config.toolmate.saveChat(config.currentMessages)
        config.currentMessages = CallLLM.resetMessages()
        self.centralWidget.contentView.setPlainText("")

    def openConversation(self):
        ...


from toolmate import isCommandInstalled, print1, print3
import os

macVlc = "/Applications/VLC.app/Contents/MacOS/VLC"
windowsVlc = r'C:\Program Files\VideoLAN\VLC\vlc.exe'

if not isCommandInstalled("vlc") and not os.path.isfile(macVlc) and not os.path.isfile(windowsVlc):

    checks = {
        "pkg": "pkg install vlc", # Android Termux
        "brew": "brew install vlc", # on MacOS using Homebrew (https://brew.sh/)
        "choco": "choco install vlc", # on Windows using Chocolatey (https://chocolatey.org/)
        "scoop": "scoop install vlc", # on Windows using Scoop (https://scoop.sh/)
        "apt": "sudo apt update && sudo apt install vlc", # Ubuntu or Debian-based distributions
        "dnf": "sudo dnf install vlc", # Fedora or CentOS
        "pacman": "sudo pacman -Sy vlc", # Arch Linux
        "zypper": "sudo zypper install vlc", # openSUSE
        "yum": "sudo yum install epel-release && sudo yum install vlc", # RHEL, CentOS 7
    }

    for command, commandline in checks.items():
        if isCommandInstalled(command):
            os.system(commandline)
            if isCommandInstalled("vlc"):
                break

    if not isCommandInstalled("vlc") and not os.path.isfile(macVlc) and not os.path.isfile(windowsVlc):
        print3("Note: 'vlc' is not installed.")
        print1("It is nice to have VLC player installed for video / audio playback. It is required if you want to control the LetMeDoIt AI audio response speed.")



"""
ToolMate AI Plugin - analyze audio file with Groq

analyze audio file with Groq

[TOOL_CALL]
"""

from toolmate import config, showErrors, getGroqClient
from toolmate import print3
import os, shutil, subprocess

# Function method
def analyze_audio_groq(function_args):
    from toolmate import config
    def check_file_format(file_path):
        # List of allowed file extensions
        allowed_extensions = ('.mp3', '.mp4', '.mpeg', '.mpga', '.m4a', '.wav', '.webm')
        # Getting the file extension
        _, file_extension = os.path.splitext(file_path)
        # Checking if the file extension is in the list of allowed extensions
        return True if file_extension.lower() in allowed_extensions else False

    audio_file = function_args.get("audio_filepath") # required

    if audio_file and os.path.isfile(audio_file):
        if shutil.which("ffmpeg"):
            temp_audio_file = os.path.join(config.toolMateAIFolder, "temp", os.path.basename(audio_file))
            if os.path.isfile(temp_audio_file):
                os.remove(temp_audio_file)
            cli = f'''ffmpeg -i "{audio_file}" -ar 16000 -ac 1 -map 0:a: "{temp_audio_file}"'''
            run_cli = subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            *_, stderr = run_cli.communicate()
            if not stderr:
                audio_file = temp_audio_file
        if not check_file_format(audio_file):
            print3("This feature supports the following input file types only: '.mp3', '.mp4', '.mpeg', '.mpga', '.m4a', '.wav', '.webm'!")
            return ""
        elif os.path.getsize(audio_file) / (1024*1024) > 25:
            print3("Audio files are currently limited to 25 MB!")
            return ""
        try:
            # read https://console.groq.com/docs/speech-text
            with open(audio_file, "rb") as file:
                transcription = getGroqClient().audio.transcriptions.create(
                    file=(audio_file, file.read()),
                    model="whisper-large-v3",
                    #prompt="Specify context or spelling",  # Optional
                    #response_format="json",  # Optional
                    #language="en",  # Optional
                    temperature=0.0  # Optional
                )
                return transcription.text
        except:
            showErrors()
        return ""

    return "[INVALID]"

# Function Signature
functionSignature = {
    "examples": [
        "analyze speech",
    ],
    "name": "analyze_audio_groq",
    "description": f'''Retrieve information from an audio with Groq''',
    "parameters": {
        "type": "object",
        "properties": {
            "audio_filepath": {
                "type": "string",
                "description": "Return the audio file path that I specified in my requests. Return an empty string '' if it is not specified.",
            },
        },
        "required": ["audio_filepath"],
    },
}

# Integrate the signature and method into LetMeDoIt AI
config.addFunctionCall(signature=functionSignature, method=analyze_audio_groq)
